@@ -1,9 +1,8 @@
 /*= -*- c-basic-offset: 4; indent-tabs-mode: nil; -*-
  *
  * librsync -- generate and apply network deltas
- * $Id: trace.h,v 1.10 2002/05/15 05:09:48 mbp Exp $
  * 
- * Copyright (C) 2000, 2001, 2002 by Martin Pool <mbp@samba.org>
+ * Copyright (C) 2000, 2001, 2002, 2003, 2004 by Martin Pool
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU Lesser General Public License as published by
@@ -24,28 +23,55 @@
 /**
  * @file
  *
+ * Reusable trace library.
+ *
  * @todo A function like perror that includes strerror output.  Apache
  * does this by adding flags as well as the severity level which say
  * whether such information should be included.
+ *
+ * @todo Also check in configure for the C9X predefined identifier `_function', or
+ * whatever it's called.
+ **/
+
+/* Provide simple macro statement wrappers (adapted from glib, and originally from Perl):
+ *  RS_STMT_START { statements; } RS_STMT_END;
+ *  can be used as a single statement, as in
+ *  if (x) RS_STMT_START { ... } RS_STMT_END; else ...
+ *
+ *  For gcc we will wrap the statements within `({' and `})' braces.
+ *  For SunOS they will be wrapped within `if (1)' and `else (void) 0',
+ *  and otherwise within `do' and `while (0)'.
  */
+#if !(defined (RS_STMT_START) && defined (RS_STMT_END))
+#  if defined (__GNUC__) && !defined (__STRICT_ANSI__) && !defined (__cplusplus)
+#    define RS_STMT_START	(void)(
+#    define RS_STMT_END		)
+#  else
+#    if (defined (sun) || defined (__sun__))
+#      define RS_STMT_START	if (1)
+#      define RS_STMT_END	else (void)0
+#    else
+#      define RS_STMT_START	do
+#      define RS_STMT_END	while (0)
+#    endif
+#  endif
+#endif
+
 
 #include <stdarg.h>
 
-/*
- * trace may be turned off.
- *
- * error is always on, but you can return and continue in some way
- *
- * fatal terminates the whole process
- */
-
+/* unconditionally on */
+#define DO_RS_TRACE
 
 /**
- * \brief Log severity levels.
+ * Log severity levels.
  *
- * These are the same as syslog, at least in glibc.
+ * These have the same numeric values as the levels for syslog, at
+ * least in glibc.
  *
- * \sa rs_trace_set_level()
+ * Trace may be turned off.
+ *
+ * Error is always on, but you can return and continue in some way.
  */
 typedef enum {
     RS_LOG_EMERG         = 0,   /**< System is unusable */
@@ -58,47 +84,61 @@ typedef enum {
     RS_LOG_DEBUG         = 7    /**< Debug-level messages */
 } rs_loglevel;
 
+int rs_loglevel_from_name(const char *name);
+
+enum {
+    RS_LOG_PRIMASK       = 7,   /**< Mask to extract priority
+                                   part. \internal */
+
+    RS_LOG_NONAME        = 8,   /**< \b Don't show function name in
+                                   message. */
+
+    RS_LOG_NO_PROGRAM   = 16,
+    RS_LOG_NO_PID       = 32
+};
 
 
 /**
- * \typedef rs_trace_fn_t
+ * \typedef rs_logger_fn
  * \brief Callback to write out log messages.
  * \param level a syslog level.
  * \param msg message to be logged.
+ *
+ * \param private Opaque data passed in when logger was added.  For
+ * example, pointer to file descriptor.
  */
-typedef void    rs_trace_fn_t(int flags, const char *fn,
-                              char const *msg, va_list);
+typedef void    rs_logger_fn(int flags, const char *fn,
+                              char const *msg, va_list,
+                              void *private_ptr, int private_int);
 
 void rs_format_msg(char *buf, size_t, int, const char *,
                    const char *fmt, va_list);
 
 void            rs_trace_set_level(rs_loglevel level);
 
-/** Set trace callback. */
-void            rs_trace_to(rs_trace_fn_t *);
-
-/** Default trace callback that writes to stderr.  Implements
- * ::rs_trace_fn_t, and may be passed to rs_trace_to(). */
-void rs_trace_stderr(int level, const char *fn, char const *fmt, va_list va);
-int rs_trace_fd;
+void rs_add_logger(rs_logger_fn *, int level, void *, int);
+void rs_remove_logger(rs_logger_fn *, int level, void *, int);
+void rs_remove_all_loggers(void);
 
 
-void
-rs_trace_syslog(int level, const char *fn, char const *fmt, va_list va);
+void rs_logger_file(int level, const char *fn, char const *fmt, va_list va,
+                    void *, int);
 
-/** Check whether the library was compiled with debugging trace
- * suport. */
+void rs_logger_syslog(int level, const char *fn, char const *fmt, va_list va,
+                      void *, int);
+
+/** Check whether the library was compiled with debugging trace suport. */
 int             rs_supports_trace(void);
 
-
-void rs_fatal0(char const *s, ...);
-void rs_error0(char const *s, ...);
-void rs_trace0(char const *s, ...);
-
-void rs_log0_nofn(int level, char const *fmt, ...);
-
 void rs_log0(int level, char const *fn, char const *fmt, ...)
-    __attribute__ ((format(printf, 3, 4)));
+#if defined(__GNUC__)
+    __attribute__ ((format(printf, 3, 4)))
+#endif /* __GNUC__ */
+  ;
+
+
+  /* TODO: Check for the __FUNCTION__ thing, rather than gnuc */
+#if defined(HAVE_VARARG_MACROS)  && defined(__GNUC__)
 
 #if 1 || defined(DO_RS_TRACE)
 #  define rs_trace(fmt, arg...)                            \
@@ -108,21 +148,12 @@ void rs_log0(int level, char const *fn, char const *fmt, ...)
 #  define rs_trace(s, str...)
 #endif	/* !DO_RS_TRACE */
 
-/*
- * TODO: Don't assume this is a gcc thing; rather test in autoconf for
- * support for __FUNCTION__ and varargs macros.  One simple way might
- * just be to try compiling the definition of one of these functions!
- *
- * TODO: Also look for the C9X predefined identifier `_function', or
- * whatever it's called.
- */
-
 #define rs_log(l, s, str...) do {              \
      rs_log0((l), __FUNCTION__, (s) , ##str);  \
      } while (0)
 
 
-#define rs_log_critical(s, str...) do {                         \
+#define rs_log_crit(s, str...) do {                         \
      rs_log0(RS_LOG_CRIT,  __FUNCTION__, (s) , ##str);          \
      } while (0)
 
@@ -142,24 +173,32 @@ void rs_log0(int level, char const *fn, char const *fmt, ...)
      rs_log0(RS_LOG_INFO,  __FUNCTION__, (s) , ##str);          \
      } while (0)
 
+#else /* not defined HAVE_VARARG_MACROS */
 
-#define rs_fatal(s, str...) do {               \
-     rs_log0(RS_LOG_CRIT,  __FUNCTION__,       \
-	      (s) , ##str);                    \
-     abort();                                  \
-     } while (0)
+/* If we don't have gcc vararg macros, then we fall back to making the
+ * log routines just plain functions.  On platforms without gcc (boo
+ * hiss!) this means at least you get some messages, but not the nice
+ * function names etc. */
+#define rs_log rs_log0_nofn
+
+#define rs_trace        rs_log_trace_nofn
+#define rs_log_info     rs_log_info_nofn
+#define rs_log_notice   rs_log_notice_nofn
+#define rs_log_warning  rs_log_warning_nofn
+#define rs_log_error    rs_log_error_nofn
+#define rs_log_crit     rs_log_critical_nofn
+#endif /* HAVE_VARARG_MACROS */
 
 
-enum {
-    RS_LOG_PRIMASK       = 7,   /**< Mask to extract priority
-                                   part. \internal */
 
-    RS_LOG_NONAME        = 8,   /**< \b Don't show function name in
-                                   message. */
+void rs_log_trace_nofn(char const *s, ...);
+void rs_log_info_nofn(char const *, ...);
+void rs_log_notice_nofn(char const *, ...);
+void rs_log_warning_nofn(char const *s, ...);
+void rs_log_error_nofn(char const *s, ...);
+void rs_log_critical_nofn(char const *, ...);
 
-    RS_LOG_NO_PROGRAM   = 16,
-    RS_LOG_NO_PID       = 32
-};
+void rs_log0_nofn(int level, char const *fmt, ...);
 
 
 
@@ -177,3 +216,11 @@ extern int rs_trace_level;
 #else
 #  define rs_trace_enabled() 0
 #endif
+
+/**
+ * Name of the program, to be included in log messages.
+ *
+ * @note This must be defined exactly once in each program that links to
+ * trace.c
+ **/
+extern const char *rs_program_name;
